#define INCL_WIN
#define INCL_DOS
#define INCL_DOSERRORS
#include <os2.h>

#include <stdio.h>
#include <stddef.h>
#include <getopt.h>
#include <stdarg.h>
#include <string.h>

#define NETSCAPE_APP "NETSCAPE"
#define TOPIC_OPENURL "WWW_OpenURL"
#define TOPIC_ACTIVATE "WWW_Activate"
#define NSCLIENT "NSCLIENT"
#define NETSCAPE_CMD "netscape.exe"
int execute_delay = 15;
int executed_poling_interval = 3;
int execute_retry_max = 5;

HAB hab;

HWND hFrameWnd, hListWnd, hServerWnd;
PID ServerPid;
TID ServerTid;

typedef struct {
  DDESTRUCT dde;
  char item[1];
} ItemStruct;

typedef struct {
  const char *topic;
  BOOL *doit;
  char *arg;
  BOOL (*check_data)(HWND, PDDESTRUCT);
} NSAction;

CONVCONTEXT context;

BOOL true = TRUE;
BOOL false = FALSE;
BOOL activate = FALSE;
ULONG wnd_id = 0xffffffff;
char *output_file;
int exitrc = 0;

static BOOL check_open(HWND, PDDESTRUCT);
static BOOL check_activate(HWND, PDDESTRUCT);

NSAction OpenURL = { TOPIC_OPENURL, &true, NULL, check_open };
NSAction Activate = { TOPIC_ACTIVATE, &activate, NULL, check_activate };

NSAction *action_list[] = {
  &OpenURL, &Activate, NULL
};

NSAction **currentAction;

static char argbuff[1024];
char cmd_buffer[1024];

static volatile void
usage_and_die(void)
{
  fprintf(stderr, "Usage: nsclient [-an] [-o filename] [-i interval] [-e cmd] [-d delay] [-r retry] [-w window] [URL]\n");
  
  exit(1);
}

static void
exit_loop(HWND hwnd, int exitcode)
{
  if (hServerWnd)
    WinDdePostMsg(hServerWnd, hwnd, WM_DDE_TERMINATE, NULL, DDEPM_RETRY);
  WinPostMsg(hwnd, WM_CLOSE, 0, 0);
  exitrc = exitcode;
}
  
static BOOL
check_open(HWND hwnd, PDDESTRUCT dde)
{
  ULONG nswnd;
  BOOL rc = TRUE;

  nswnd = *(ULONG *)((char *)dde + dde->offabData);
#if DEBUG
  fprintf(stderr, "status = 0x%x\n", dde->fsStatus);
  fprintf(stderr, "cbData = %lu\n", dde->cbData);
  fprintf(stderr, "offszItemName = %lu\n", dde->offszItemName);
  fprintf(stderr, "offabData = %lu\n", dde->offabData);
  fprintf(stderr, "itemName = %s\n", (char *)dde + dde->offszItemName);
  fprintf(stderr, "wnd = 0x%x\n", nswnd);
#endif

  if (nswnd == 0 || dde->fsStatus != DDE_FRESPONSE) {
    fprintf(stderr, "DDE Protocol error for open URL\n");
    rc = FALSE;
  }
  DosFreeMem(dde);

  return rc;
}

static BOOL
check_activate(HWND hwnd, PDDESTRUCT dde)
{
  ULONG nswnd;
  BOOL rc = TRUE;

  nswnd = *(ULONG *)((char *)dde + dde->offabData);
#if DEBUG
  fprintf(stderr, "status = 0x%x\n", dde->fsStatus);
  fprintf(stderr, "cbData = %lu\n", dde->cbData);
  fprintf(stderr, "offszItemName = %lu\n", dde->offszItemName);
  fprintf(stderr, "offabData = %lu\n", dde->offabData);
  fprintf(stderr, "itemName = %s\n", (char *)dde + dde->offszItemName);
  fprintf(stderr, "wnd = 0x%x\n", nswnd);
#endif

  if (nswnd == 0 || dde->fsStatus != DDE_FRESPONSE) {
    fprintf(stderr, "DDE Protocol error for activate\n");
    rc = FALSE;
  }
  DosFreeMem(dde);

  return rc;
}

static char *
xsprintf(char *fmt,...)
{
  va_list argptr;
  va_start(argptr, fmt);
  vsprintf(argbuff, fmt, argptr);
  va_end(argptr);
  return strdup(argbuff);
}

static BOOL
execute_navigator(void)
{
  static BOOL executed = FALSE;
  STARTDATA start;
  ULONG rc;
  char fail_buffer[256];
  ULONG csid;
  PID cpid;

  if (executed) {
    sleep(executed_poling_interval);
  } else {
    memset(&start, 0, sizeof(start));
    start.Length = sizeof(start);
    start.Related = SSF_RELATED_INDEPENDENT;
    start.FgBg = SSF_FGBG_BACK;
    start.TraceOpt = SSF_TRACEOPT_NONE;
    start.PgmTitle = NULL;
    start.PgmInputs = cmd_buffer;
    start.TermQ = NULL;
    start.Environment = NULL;
    start.InheritOpt = SSF_INHERTOPT_SHELL;
    start.SessionType = SSF_TYPE_DEFAULT;
    start.PgmHandle = NULLHANDLE;
    /*  start.PgmControl = SSF_CONTROL_VISIBLE; */
    start.ObjectBuffer = fail_buffer;
    start.ObjectBuffLen = sizeof(fail_buffer);
    if ((rc = DosStartSession(&start, &csid, &cpid)) != NO_ERROR) {
      if (rc != ERROR_SMG_START_IN_BACKGROUND) {
	fprintf(stderr, "DosStartSession failed: rc = %lu\n", rc);
	return FALSE;
      }
    }
    executed = TRUE;
    sleep(execute_delay);
  }

  return TRUE;
}

static void
skip_action(void)
{
  while (*currentAction && !*(*currentAction)->doit)
    currentAction++;
}

static void
next_action(void)
{
  if (*currentAction == NULL)
    return;
  currentAction++;
  skip_action();
}

static void
post_request(HWND hwnd)
{
  ULONG len;
  PVOID mem;
  ItemStruct *pdde;

  if ((*currentAction)->arg == NULL) {
    len = 1;
  } else {
    len = strlen((*currentAction)->arg) + 1;
  }
  DosAllocSharedMem(&mem,
		    NULL,
		    sizeof(ItemStruct) + len,
		    PAG_COMMIT | PAG_READ | PAG_WRITE | OBJ_GIVEABLE);
  DosGiveSharedMem(&mem, ServerPid, PAG_READ | PAG_WRITE);
  pdde = (ItemStruct *)mem;
  pdde->dde.fsStatus = 0;
  pdde->dde.usFormat = DDEFMT_TEXT;
  pdde->dde.offszItemName = offsetof(ItemStruct, item);
  pdde->dde.offabData = sizeof(ItemStruct) + len;
  if ((*currentAction)->arg == NULL) {
    pdde->item[0] = 0;
  } else {
    strcpy(pdde->item, (*currentAction)->arg);
  }
  pdde->dde.cbData = 0;
  WinDdePostMsg(hServerWnd, hwnd, WM_DDE_REQUEST, (PDDESTRUCT)pdde, 0);
  if ((*currentAction)->check_data == NULL) {
    next_action();
    if (*currentAction == NULL)
      exit_loop(hwnd, 0);
  }
}

static void
initiate_topic(HWND hwnd)
{
  int retry = 0;

  while (retry < execute_retry_max) {
    BOOL rc;

    rc = WinDdeInitiate(hwnd, NETSCAPE_APP,
			(*currentAction)->topic, &context);
    if (rc && hServerWnd) {
      break;
    } else {
      if (!execute_navigator()) {
	exit_loop(hwnd, 1);
	break;
      }
    }
    retry++;
  }
  if (retry == execute_retry_max) {
    fprintf(stderr, "DDE cannot initiated\n");
    exit_loop(hwnd, 1);
  }
}  

MRESULT EXPENTRY
nsc_wnd_proc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
  BOOL handled = TRUE;
  switch (msg) {
   case WM_CREATE:
   {
     if (*currentAction == NULL) {
       exit_loop(hwnd, 0);
       break;
     }
     initiate_topic(hwnd);
     /* handled = FALSE; */
   }
   break;
   case WM_DDE_INITIATEACK:
   {
     PDDEINIT init;

#if DEBUG
     fprintf(stderr, "initiateack\n");
#endif
     hServerWnd = (HWND)mp1;
     WinQueryWindowProcess(hServerWnd, &ServerPid, &ServerTid);
     init = (PDDEINIT)mp2;
     if (stricmp(init->pszTopic, (*currentAction)->topic) == 0) {
       post_request(hwnd);
     }
   }
   break;
   case WM_DDE_DATA:
   {
     PDDESTRUCT ddestruct = mp2;

#if DEBUG
     fprintf(stderr, "data\n");
#endif
     if (ddestruct) {
       APIRET rc;
       if ((rc = DosGetSharedMem((PVOID)ddestruct, PAG_READ|PAG_WRITE))) {
	 fprintf(stderr, "DosGetSharedMem failed, rc = %lu\n", rc);
	 exit_loop(hwnd, 1);
	 break;
       }
     }
     if (*currentAction && (*currentAction)->check_data) {
       if (!(*(*currentAction)->check_data)(hwnd, ddestruct)) {
	 exit_loop(hwnd, 1);
       }
     }
     WinDdePostMsg(hServerWnd, hwnd, WM_DDE_TERMINATE, NULL, DDEPM_RETRY);
     hServerWnd = 0;
     next_action();
     if (*currentAction == NULL)
       exit_loop(hwnd, 0);
     else
       initiate_topic(hwnd);
   }
   break;
   default:
#if DEBUG
    fprintf(stderr, "msg = 0x%x\n", msg);
#endif
    handled = FALSE;
    break;
  }
  if (!handled)
    return WinDefWindowProc(hwnd, msg, mp1, mp2);
  return 0;
}

int
main(int argc, char **argv)
{
  HMQ hmq;
  FRAMECDATA fcd;
  QMSG qmsg;
  BOOL rc;
  int c;
  PTIB  ptib;
  PPIB  ppib;

  strcpy(cmd_buffer, "/c " NETSCAPE_CMD);
  while ((c = getopt(argc, argv, "o:ai:ne:d:r:w:")) != EOF) {
    switch (c) {
     case 'o':
      output_file = optarg;
      break;
     case 'a':
      activate = TRUE;
      break;
     case 'w':
      wnd_id = atol(optarg);
      break;
     case 'n':
      wnd_id = 0;
      break;
     case 'e':
      strcpy(cmd_buffer + 3, optarg);
      break;
     case 'd':
      execute_delay = atol(optarg);
      break;
     case 'r':
      execute_retry_max = atol(optarg);
      break;
     case 'i':
      executed_poling_interval = atol(optarg);
      break;
     default:
      usage_and_die();
      break;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc == 0) {
    OpenURL.doit = &false;
  } else {
    
    OpenURL.arg = xsprintf("%s,%s,0x%x", argv[0],
			   output_file ? output_file : "", wnd_id);
  }
  Activate.arg = xsprintf("0x%x", wnd_id);

  currentAction = action_list;
  context.cb = sizeof(context);

  skip_action();

  DosGetInfoBlocks(&ptib, &ppib);
  ppib->pib_ultype = PROG_PM;

  if (!(hab = WinInitialize(0)))
    return 1;
  if (!(hmq = WinCreateMsgQueue(hab, 0)))
    return 1;

  if (!WinRegisterClass(hab, NSCLIENT, (PFNWP)nsc_wnd_proc, 0, 0))
    return 1;

  fcd.cb = sizeof(fcd);
  fcd.flCreateFlags = 0;
  fcd.hmodResources = NULLHANDLE;
  fcd.idResources = 0;
  hFrameWnd = WinCreateWindow(HWND_DESKTOP,
			      NSCLIENT,
			      "Netscape Client",
			      0, 0, 0, 0, 0,
			      NULLHANDLE,
			      HWND_TOP,
			      1000,
			      &fcd,
			      NULL);
  if (!hFrameWnd)
    return 1;
  hListWnd = WinCreateWindow(hFrameWnd,
			     WC_LISTBOX,
			     NULL,
			     LS_HORZSCROLL,
			     0, 0, 0, 0,
			     hFrameWnd,
			     HWND_BOTTOM,
			     FID_CLIENT,
			     NULL,
			     NULL);

  WinShowWindow(hFrameWnd, TRUE);

  while (WinGetMsg(hab, &qmsg, NULLHANDLE, 0, 0)) {
    WinDispatchMsg(hab, &qmsg);
  }

  WinDestroyMsgQueue(hmq);
  WinTerminate(hab);

  return exitrc;
}
